/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_BIT_SPLIT_VISITOR__
#define __DD_BIT_SPLIT_VISITOR__

#include <vector>

#include "DDBitList.h"
#include "DDRayRefVisitor.h"

#include "DDRayReferenceContainer.h"
#include "DDTreeNodeCondition.h"
#include "DDExtendedBitPatternCondition.h"
#include "DDVantagePointTreeNodeCondition.h"
#include "DDTwoBitsFragmentationCondition.h"
#include "DDFragmentationCondition.h"

namespace addibit {

    enum BitSplitVisitorState {
        TERMINATE       = 0x00000001,
        CHANGE_VISITOR  = 0x00000002,
        CONTINUE        = 0x00000004
    };

    enum ConditionSupport {
        BIT_PATTERN,
        VANTAGE_POINT,
        POPULATION
    };

    class BitSplitVisitor : public RayRefVisitor {
    
        private:
        uint_fast32_t mContainerIDCounter;

        protected:
        uint_fast16_t mCurrentDepth;

        const uint_fast16_t mMaxDepth;
        const uint_fast16_t mLeafCapacity;

        RayReferenceContainer * mSets;
        unsigned int mSetsCount;

        BitSplitVisitor * mSuccessor;
  
        unsigned long mEntriesCount;

        RayReferenceContainer & set ( unsigned int index ) {
            return mSets[ index ];
        }

        const RayReferenceContainer & set ( unsigned int index ) const {
            return mSets[ index ];
        }
        
        virtual void registerContainer ( const unsigned int ) {}
        virtual unsigned int getTargetContainer ( const RayReference & ) = 0;

        public:
        BitSplitVisitor( const unsigned int max_depth,
                         const unsigned int capacity,
                         const unsigned int split ) :
            RayRefVisitor(),
            mCurrentDepth( 0 ),
            mMaxDepth( max_depth ),
            mLeafCapacity( capacity ),
            mSetsCount( split ),
            mSuccessor( NULL ) {}

        virtual ~BitSplitVisitor () {
            if ( hasSuccessor() ) {
                delete mSuccessor;
            }
        }

        virtual void init ( const RayReferenceContainer & ) {
            mContainerIDCounter = 0;    
            registerContainer( mContainerIDCounter++ );    
        }

        virtual void onStart () {}

        void onVisit( const RayReference & ref ) {}
        
        void onPop ( RayReference && ref ) {
            const uint_fast16_t target = getTargetContainer( ref );
            set( target ).insert( std::move( ref ) );
        }

        void registerOutputSets ( RayReferenceContainer * sets ) {
            mSets = sets;
            
            for ( uint_fast16_t s_index = 0; s_index < mSetsCount; ++s_index ) {
                set( s_index ).setId( mContainerIDCounter );
                registerContainer( mContainerIDCounter++ );
            }            
        }

        virtual void before ( const RayReferenceContainer & container ) {
            mEntriesCount = container.size();
            mCurrentDepth++;
        }

        virtual void after () {
            mCurrentDepth--;
        }      

        virtual bool can_proceed ( const unsigned int ) const {
            return true;
        }

        virtual bool can_proceed ( const RayReferenceContainer & ) const {
            return true;
        }

        unsigned int getSplitFactor () const {
            return mSetsCount;
        }

        void setSuccessor ( BitSplitVisitor * successor ) {
            mSuccessor = successor;
        }

        BitSplitVisitor * successor () {
            return mSuccessor;
        }

        const BitSplitVisitor * successor () const {
            return mSuccessor;
        }

        bool hasSuccessorToProceed ( const RayReferenceContainer & container ) const {
            return hasSuccessor() && ( successor()->can_proceed( container ) || successor()->hasSuccessorToProceed( container ));
        }

        BitSplitVisitor * successor_to_proceed ( const RayReferenceContainer & container ) {
            if ( successor()->can_proceed( container ) ) {
                return successor();
            } else {
                return successor()->successor_to_proceed( container );
            }
        }

        bool hasSuccessor() const {
            return mSuccessor != NULL ;
        }

        inline bool leaf_condition ( const unsigned int index ) const {
            return mCurrentDepth >= mMaxDepth || set( index ).size() <= mLeafCapacity;
        }
        
        inline bool leaf_condition ( const RayReferenceContainer & container ) const {
            return container.size() <= mLeafCapacity;
        }        

        void setCurrentDepth ( const unsigned int depth ) {
            mCurrentDepth = depth;
        }

        unsigned int current_depth () const {
            return mCurrentDepth;
        }

        BitSplitVisitorState state ( const unsigned int index ) const {
            const bool proceed = can_proceed( index );

            if ( leaf_condition( index ) || unlikely( !proceed && !hasSuccessorToProceed( set( index ) ) ) ) {
                return TERMINATE;
            } else if ( !proceed ) {
                return CHANGE_VISITOR;
            } else {
                return CONTINUE;
            }
        }
        
        inline uint_fast16_t max_depth () const {
            return mMaxDepth;
        }
    };
}

#endif // __DD_BIT_SPLIT_VISITOR__
