/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_CACHE_FACTORY__
#define __DD_CACHE_FACTORY__

#include "DDTerminalCacheEntryNode.h"
#include "DDBasisCacheEntryNode.h"

namespace addibit {

    template <class RAY_TYPE, class CACHE_ENTRY_TYPE>
    class CacheFactory {
    
        private:
        std::deque<std::shared_ptr<const RAY_TYPE > > * const mExtremalRays;
        const uint_fast16_t mThreads;
        
        uint_fast16_t mColumns;
        uint_fast16_t mListLikeRows;
        uint_fast16_t mTotalRows;
        std::vector<uint_fast16_t> __strict_inequalities;
        
        uint_fast16_t mId;
        
        BitList mPattern;
        
	    inline CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * create_list_like_cache ( CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * anchor ) {
	        
		    CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * current_node = anchor;
		    CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * next_node = NULL;
		            
			for ( uint_fast16_t index = mColumns; index < mColumns + mListLikeRows - 1; index++, current_node = next_node ) {
						    
			    next_node = new IntermediateCacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE>( CACHE_ENTRY_TYPE::inequality_entry( mTotalRows, index, mPattern ), 1, mId++ );		    
			    current_node->insert( next_node, 0 );
                
                mPattern.setBit( index );
		    }	     
		    
		    if ( mListLikeRows + mColumns == mTotalRows ) {                 
		        next_node = new TerminalCacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE>( CACHE_ENTRY_TYPE::inequality_entry( mTotalRows, mTotalRows - 1, mPattern ), mExtremalRays[0], __strict_inequalities, mId );
		    } else {
		        next_node = new IntermediateCacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE>( CACHE_ENTRY_TYPE::inequality_entry( mTotalRows, mColumns + mListLikeRows - 1, mPattern ), mTotalRows - ( mColumns + mListLikeRows ), mId++ );
		    }         
		    current_node->insert( next_node, 0 );
		    mPattern.setBit( mColumns + mListLikeRows - 1 );
            
		    return next_node;                                                             
	    }
        
        inline void create_tree_like_cache ( CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * anchor,
                                             const uint_fast16_t tree_root_row,
                                             const uint_fast16_t starting_row,
                                             const uint_fast16_t threads ) {
        
            const Interval thread_pool( 0, threads - 1 );
            create_tree_like_cache( anchor, tree_root_row, starting_row, __strict_inequalities, thread_pool );
        }
	    
        void create_tree_like_cache ( CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * anchor,
                                      const uint_fast16_t tree_root_row,
                                      const uint_fast16_t starting_row,
                                      const std::vector<uint_fast16_t> & previous_strict_inequalities,
                                      const Interval & pool ) {
            
            std::vector<uint_fast16_t> strict_inequalities = previous_strict_inequalities;
            
            Interval thread_id_pool( pool );
        
            for ( uint_fast16_t index = starting_row; index < mTotalRows - 1; index++ ) {
                const Interval next_pool = next_thread_pool( thread_id_pool );
                
                CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * next_node = new IntermediateCacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE>( CACHE_ENTRY_TYPE::equality_entry( mTotalRows, index, tree_root_row, mPattern ), mTotalRows - ( index + 1 ), mId++, next_pool.lb() );
                anchor->insert( next_node, index - starting_row );
                
                mPattern.setBit( index );
                create_tree_like_cache( next_node, tree_root_row, index + 1, strict_inequalities, next_pool );
                mPattern.unsetBit( index );
                
                strict_inequalities.push_back( index );
            }
            anchor->insert( new TerminalCacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE>( CACHE_ENTRY_TYPE::inequality_entry( mTotalRows, mTotalRows - 1, tree_root_row, mPattern ), mExtremalRays[ thread_id_pool.lb() ], strict_inequalities, mId++, thread_id_pool.lb() ), mTotalRows - ( starting_row + 1 ) );
	    }  
        
        const Interval next_thread_pool ( Interval & current_pool ) const {
            
            if ( likely( current_pool.pointed() ) ) {
                return current_pool;
            } else {
                const Interval result( current_pool.lb() + current_pool.range() / 2, current_pool.ub() );
                current_pool.ub() -= current_pool.range() / 2; 
                
                return result;
            }
        }  
        
        inline void init ( const uint_fast16_t columns,
	                       const uint_fast16_t list_like_rows,
	                       const uint_fast16_t total_rows ) {
            mColumns = columns;
            mListLikeRows = list_like_rows;
            mTotalRows = total_rows;
                            
            for ( uint_fast16_t i = mTotalRows; i < GLOBAL_MATRIX_ROWS; ++i ) {
                __strict_inequalities.push_back( i );
            }
            
            mId = 1;
            
            mPattern.setBits( 0, columns );
        }
	    
        public:
        CacheFactory ( std::deque<std::shared_ptr<const RAY_TYPE > > * extremal_rays,
                       const uint_fast16_t threads ) :
            mExtremalRays( extremal_rays ),
            mThreads( threads ),
            mPattern( __BIT_LIST_ALLOC_SPACE__, 0 ) {}
        
	    BasisCacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * create_cache ( const uint_fast16_t columns,
                                                                         const uint_fast16_t list_like_rows,
                                                                         const uint_fast16_t total_rows ) {
	    
            init( columns, list_like_rows, total_rows );
            
		    BasisCacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * anchor = new BasisCacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE>( 1 );
		    CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * last_list_entry = create_list_like_cache( anchor );
            
            if ( list_like_rows + columns < total_rows ) {
                create_tree_like_cache( last_list_entry, columns + list_like_rows, columns + list_like_rows, mThreads );
            }
		    
		    return anchor;
	    }         
    };
}

#endif /* __DD_CACHE_FACTORY__ */
