/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_CONDITIONED_TREE_NODE__
#define __DD_CONDITIONED_TREE_NODE__

#include "DDTreeNode.h"
#include "DDRayRefContainerToBitListArrayTransformer.h"

namespace addibit {
    
    class ConditionedTreeNode {
    
        public:
        ConditionedTreeNode * __next_if_cond_true;
        ConditionedTreeNode * __next_if_cond_false;
        
    #ifdef __OUTPUT_TREES__
        ConditionedTreeNode * __parent;
        std::string __label;
    #endif
        
        BitListArray mArray;
        uintptr_t * mRefId;
        
        __BIT_LIST_RAW_DATA_TYPE__ __data[ 1 ];
        
        void __attribute__((always_inline)) narrow_ref ( EdgeCandidatesContainer & container ) const {      
                
            const __BIT_LIST_RAW_DATA_TYPE__ * __entry = mArray.data_begin();
                    
            while ( mArray.next_feasible( container.candidate(), __entry ) ) {
                container.push( __entry, mRefId[ ( __entry - mArray.data_begin() ) / __DD_BLOCK_COUNT__ ] );
                __entry += __DD_BLOCK_COUNT__;
            }
        }    
        
        bool __attribute__((always_inline)) verify_ref ( const EdgeCandidate & candidate ) const {   
            
            int_fast32_t hit_index = -1;
            
            do {
                hit_index = mArray.contains( candidate.data(), ++hit_index );
            } while ( hit_index < mArray.entries() && ( mRefId[ hit_index ] == candidate.illegal_1() || mRefId[ hit_index ] == candidate.illegal_2() ) );
            
            return hit_index < mArray.entries(); 
        }              
        
        public:
        ConditionedTreeNode () :
            __next_if_cond_true( NULL ),
            __next_if_cond_false( NULL ) {}          
        
        ConditionedTreeNode ( const RayReferenceContainer & candidates,
                              char * __elements_segment ) :
            __next_if_cond_true( NULL ),
            __next_if_cond_false( NULL ),
            mArray( candidates.size(), __elements_segment ),
            mRefId( new uintptr_t[ candidates.size() ] ) {
                
            RayRefContainerToBitListArrayTransformer transformer( mArray, mRefId );
            candidates.iterate( transformer );                
        }        
        
        ConditionedTreeNode ( const RayReferenceContainer & candidates,
                              char * __maps_segment,
                              uintptr_t * __ids_segment ) :
            __next_if_cond_true( NULL ),
            __next_if_cond_false( NULL ),
            mArray( candidates.size(), __maps_segment ),
            mRefId( __ids_segment ) {
                
            RayRefContainerToBitListArrayTransformer transformer( mArray, mRefId );
            candidates.iterate( transformer );                
        }  
        
        inline __BIT_LIST_RAW_DATA_TYPE__ * data_segment () {
            return __data;
        }
        
        template<class COND_TYPE>
		void narrow ( EdgeCandidatesContainer & container ) const {
                          
            const ConditionedTreeNode * __current = this;

            do { 
                if ( COND_TYPE::narrow_feasible( container, __current->__data ) ) {
                    if ( __current->mArray.empty() ) { /* intermediate*/
                        COND_TYPE::intermediate_post_action( container, __current->__data );
                    } else {
                        __current->narrow_ref( container );
                        COND_TYPE::leaf_post_action( container, __current->__data );
                    }
                    __current = __current->__next_if_cond_true;
                } else {
                    __current = __current->__next_if_cond_false;
                } 
            } while ( __current != NULL );    
        } 
        
        template<class COND_TYPE>
		bool verify ( EdgeCandidate & candidate ) const {

            const ConditionedTreeNode * __current = this;
            bool redundant = false;
            
            do {
                if ( COND_TYPE::verify_feasible( candidate, __current->__data ) ) {
                    if ( __current->mArray.empty() ) { /* intermediate*/
                        redundant = COND_TYPE::intermediate_post_action( candidate, __current->__data );
                    } else {
                        redundant = COND_TYPE::leaf_post_action( candidate, __current->__data ) || __current->verify_ref( candidate );
                    }
                    __current = __current->__next_if_cond_true;
                } else {
                    __current = __current->__next_if_cond_false;
                }
            } while ( !redundant && likely( __current != NULL ) );
            
            return redundant; 
        }   
        
        inline void on_cond_true ( ConditionedTreeNode * next ) {
            __next_if_cond_true = next;
        }
        
        inline void on_cond_false ( ConditionedTreeNode * next ) {
            __next_if_cond_false = next;
        }        
        
        inline void on_every_cond ( ConditionedTreeNode * next ) {
            __next_if_cond_false = next;
            __next_if_cond_true = next;
        }     
        
        inline void on_every_unset_cond ( ConditionedTreeNode * next ) {
            if ( !__next_if_cond_false ) {
                __next_if_cond_false = next;
            }
            if ( !__next_if_cond_true ) {
                __next_if_cond_true = next;
            }
        }                
        
        inline const ConditionedTreeNode * on_cond_true () const {
            return __next_if_cond_true;
        }
        
        inline const ConditionedTreeNode * on_cond_false () const {
            return __next_if_cond_false;
        }        
        
        const std::string toString () const {
            return ""; //mCondition.toString();
        }
    };
}

#endif /* __DD_CONDITIONED_TREE_NODE__ */
