/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DD_EXTREME_RAY_ENUMERATOR__
#define __DD_EXTREME_RAY_ENUMERATOR__

#include <vector>
#include <sstream>
#include <algorithm>
#include <math.h>

#include "DDStats.h"

#include "DDOptions.h"

#include "DDBitList.h"
#include "DDBitCountTable.h"
#include "DDSingleSetRay.h"
#include "DDBitSelector.h"
#include "DDMatrix.h"

#include "DDTreeFactory.h"
#include "DDBitSelectionStrategyFactory.h"

#include "DDRayReferenceContainer.h"

#include "DDBitFrequencyModifier.h"
#include "DDZeroFrequencyModifier.h"

#include "DDDegeneracyCalcVisitor.h"
#include "DDRayRefReleaseVisitor.h"

#include "DDCacheEntryNode.h"
#include "DDLeafTreeCreator.h"

namespace addibit {
	
	template <class RAY_TYPE, class CACHE_ENTRY_TYPE, class POSITIVE_TRAVERSE_VISITOR, class NEGATIVE_TRAVERSE_VISITOR>
	class ExtremeRayEnumerator {
		
        private:
        const uint_fast16_t mEdgeDim;        
        
        protected:
        const TreeFactory mTreeFactory;
        const Matrix & mSystem;
        BitSelectionStrategyFactory<RAY_TYPE> mStrategyFactory;

        const uint_fast16_t mPlane;

        CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> & mCacheNode;

        template<class TRAVERSE_VISITOR>
        void run_adjacency_test ( const RayWrapperContainer<RAY_TYPE> & list_container,
                                  TRAVERSE_VISITOR traverse_visitor,
                                  StepStats & stats ) {
                                       
            list_container.template iterate<TRAVERSE_VISITOR>( traverse_visitor );
            stats.narrowed += traverse_visitor.narrowed(); 
            stats.verified += traverse_visitor.verified();
            stats.extremal += traverse_visitor.extremal();     
        }

        void run_adjacency_test ( StepStats & );
        void run_adjacency_test ( const RayWrapperContainer<RAY_TYPE> &, const RayWrapperContainer<RAY_TYPE> &, StepStats &, const bool, const TreeOptions &
                                                                                                                                                        #ifdef __OUTPUT_TREES__
                                                                                                                                                        , const char *
                                                                                                                                                        #endif
                                                                                                                                                         );         

    #ifdef __STATS_OUTPUT__        
        virtual void fill_stats ( struct StepStats & stats );
    #endif    
        
        virtual POSITIVE_TRAVERSE_VISITOR positive_traverse_visitor ( const TreeNode * const, const RayWrapperContainer<RAY_TYPE> &, const uint_fast16_t ) const = 0;
        virtual NEGATIVE_TRAVERSE_VISITOR negative_traverse_visitor ( const TreeNode * const, const RayWrapperContainer<RAY_TYPE> &, const uint_fast16_t ) const = 0;

        inline uint_fast16_t edge_dim () const {
            return mEdgeDim;
        }
        
        inline uint_fast16_t plane () const {
            return mPlane;
        }

        public:
        ExtremeRayEnumerator ( const Matrix & system,
                               const uint_fast16_t plane,
                               CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * node,
                               const uint_fast16_t containers ) :
            mEdgeDim( system.getColumns() - 2 ),
            mTreeFactory( plane ),
            mSystem( system ),
            mStrategyFactory( plane, containers ),
            mPlane( plane ),
            mCacheNode( *node ) {}

		const struct StepStats enumerate () {

            struct StepStats stats;            

            if ( likely( mCacheNode.data().getPositiveCount() > 0 ) && likely ( mCacheNode.data().getNegativeCount() > 0 ) ) {
                
                stats.degeneracy = mCacheNode.data().degeneracy();
    #ifdef __STATS_OUTPUT__ // @FIXME Multithreading problems
                TreeNodeCondition::start_new_stat_session();
    #endif
                /* This is where all the magic happens. */
                run_adjacency_test( stats );

    #ifdef __STATS_OUTPUT__
			    fill_stats( stats );
    #endif
            }

			return stats;
		}	
    };
    
    template <class RAY_TYPE, class CACHE_ENTRY_TYPE, class POSITIVE_TRAVERSE_VISITOR, class NEGATIVE_TRAVERSE_VISITOR>
    void ExtremeRayEnumerator<RAY_TYPE, CACHE_ENTRY_TYPE, POSITIVE_TRAVERSE_VISITOR, NEGATIVE_TRAVERSE_VISITOR>::run_adjacency_test ( const RayWrapperContainer<RAY_TYPE> & list_container,
                                                                                                                                      const RayWrapperContainer<RAY_TYPE> & tree_container,
                                                                                                                                      StepStats & stats,
                                                                                                                                      const bool pos_traverse_visitor,
                                                                                                                                      const TreeOptions & options
                                                                                                                                #ifdef __OUTPUT_TREES__
                                                                                                                                      , const char * label
                                                                                                                                #endif
                                                                                                                                ) {
        if ( not ( tree_container.empty() or list_container.empty() ) ) {
            const bool trivial_container = ( tree_container.size() < options.min_elements() or list_container.size() < options.min_elements() );
            AbstractTreeCreator * tree_creator;
    
    #ifdef __STATS_OUTPUT__ 
                START_TIMER( TREE_CREATION_TIME )
    #endif
    
            if ( not trivial_container ) {                
                const BitSelectStrategy * strategy = mStrategyFactory.createNarrowingStrategy( list_container,
                                                                                               tree_container,
                                                                                               options.bpt_bit_selection_strategy() );
                tree_creator = mTreeFactory.getTreeCreator( options, *strategy );
                delete strategy;
            } else {
                tree_creator = new LeafTreeCreator();
            }
            
            const TreeNode * const tree = tree_creator->createHDTree( tree_container.export_refs( mCacheNode.data().pattern() ) );
            
    #ifdef __STATS_OUTPUT__  
                STOP_TIMER( TREE_CREATION_TIME )
                stats.tree_creation_time += TIMER_VALUE( TREE_CREATION_TIME );
    #endif
            
    #ifdef __OUTPUT_TREES__
                tree->toFile( label, plane(), 6 );
    #endif     
    
            if ( pos_traverse_visitor ) {
                run_adjacency_test<POSITIVE_TRAVERSE_VISITOR>( list_container, positive_traverse_visitor( tree, tree_container, tree_creator->getAdjacencyStackSize() ), stats );
            } else {
                run_adjacency_test<NEGATIVE_TRAVERSE_VISITOR>( list_container, negative_traverse_visitor( tree, tree_container, tree_creator->getAdjacencyStackSize() ), stats );
            }   
            
    #ifdef __STATS_OUTPUT__ 
            stats.narrowing_queries += list_container.size();
            stats.narrowing_ops_linear += list_container.size() * tree_container.size();
            stats.narrowing_ops_logarithmic += list_container.size() * log( tree_container.size() );
    #endif
    
            delete tree_creator;
        }        
    }
    
    template <class RAY_TYPE, class CACHE_ENTRY_TYPE, class POSITIVE_TRAVERSE_VISITOR, class NEGATIVE_TRAVERSE_VISITOR>
    void ExtremeRayEnumerator<RAY_TYPE, CACHE_ENTRY_TYPE, POSITIVE_TRAVERSE_VISITOR, NEGATIVE_TRAVERSE_VISITOR>::run_adjacency_test ( StepStats & stats ) {
    
        const Options & options = Options::current();
    
        DD_LOG_DEBUG( "Combining Positive LD (list) and Negative HD (tree)" )
        run_adjacency_test( mCacheNode.data().positive_ld(), mCacheNode.data().negative_hd(), stats, false, options.mixed_narrowing() 
                                                                                                                                    #ifdef __OUTPUT_TREES__
                                                                                                                                    , "tree_pos_ld_to_neg_hd"
                                                                                                                                    #endif
                                                                                                                                    );
    
        DD_LOG_DEBUG( "Combining Positive LD (tree) and Negative LD (list)" )
        run_adjacency_test( mCacheNode.data().negative_ld(), mCacheNode.data().positive_ld(), stats, true, options.ld_narrowing() 
                                                                                                                                    #ifdef __OUTPUT_TREES__
                                                                                                                                    , "tree_neg_ld_to_pos_ld"
                                                                                                                                    #endif
                                                                                                                                    );
        
        DD_LOG_DEBUG( "Combining Positive HD (tree) and Negative LD (list)" )
        run_adjacency_test( mCacheNode.data().negative_ld(), mCacheNode.data().positive_hd(), stats, true, options.mixed_narrowing() 
                                                                                                                                    #ifdef __OUTPUT_TREES__
                                                                                                                                    , "tree_neg_ld_to_pos_hd"
                                                                                                                                    #endif
                                                                                                                                    );
        
        DD_LOG_DEBUG( "Combining Positive HD (tree) and Negative HD (list)" )
        run_adjacency_test( mCacheNode.data().negative_hd(), mCacheNode.data().positive_hd(), stats, true, options.hd_narrowing() 
                                                                                                                                    #ifdef __OUTPUT_TREES__
                                                                                                                                    , "tree_neg_hd_to_pos_hs"
                                                                                                                                    #endif
                                                                                                                                    );
    }
    
#ifdef __STATS_OUTPUT__        
    template <class RAY_TYPE, class CACHE_ENTRY_TYPE, class POSITIVE_TRAVERSE_VISITOR, class NEGATIVE_TRAVERSE_VISITOR>
    void ExtremeRayEnumerator<RAY_TYPE, CACHE_ENTRY_TYPE, POSITIVE_TRAVERSE_VISITOR, NEGATIVE_TRAVERSE_VISITOR>::fill_stats ( struct StepStats & stats ) {
        
        stats.narrowing_ops_actual = TreeNodeCondition::NARROWING_OPS;
        if ( stats.narrowing_ops_actual > stats.narrowing_ops_logarithmic ) {
            stats.narrowing_complexity = ( stats.narrowing_ops_actual - stats.narrowing_ops_logarithmic ) / ( stats.narrowing_ops_linear - stats.narrowing_ops_logarithmic ) + 1;
        } else {
            stats.narrowing_complexity = 1;
        }
    }	    
#endif    
}

#endif // __DD_EXTREME_RAY_ENUMERATOR__
