/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_FRAGMENTATION_TRANSFORMATION__
#define __DD_FRAGMENTATION_TRANSFORMATION__

#include <emmintrin.h>
#include <new>

#include "DDBitList.h"
#include "DDRayReference.h"
#include "DDFragmentationData.h"
#include "DDBitSelectStrategy.h"
#include "DDAdjacencyTestStack.h"

namespace addibit {

    class FragmentationTransformation {
        
        private:
        const uint_fast16_t mFragSize;
        const uint_fast16_t mFragCount;
        const uint_fast16_t __adjacency_candidate_size;
        const uint_fast16_t __stack_size;
                
        const BitList * const mFragments;
        const double_t * const __probabilities;
        
        const BitList * createFragments ( const BitSelectStrategy &, const unsigned int, const unsigned int ) const;
        BitList * const * createTransformationTable ( const unsigned int, const unsigned int ) const;    
        
        void apply ( const __BIT_LIST_RAW_DATA_TYPE__ * const map_data, __BIT_LIST_RAW_DATA_TYPE__ * aux_data ) const { 
                    
            FragmentationSegment * __segment = reinterpret_cast<FragmentationSegment *>( aux_data );
            
            if ( likely( mFragCount % 16 ) ) {
                __segment[ mFragCount / 16 ].__data = _mm_setzero_si128();
            }
            
            for ( uint_fast16_t findex = 0; findex < mFragCount; ++findex ) {
                __segment->__vec[ findex ] = mFragments[ findex ].getCommonPopCnt( map_data );
            }             
        }
        
        public:
        FragmentationTransformation ( const unsigned int fsize,
                                      const BitSelectStrategy & fstrategy ) :
            mFragSize( fsize ),
            mFragCount( ( fstrategy.getBitCount() + ( fsize - 1 )) / fsize ),
            __adjacency_candidate_size( sizeof( AdjacencyCandidate ) + sizeof( __BIT_LIST_RAW_DATA_TYPE__ ) * ( __DD_BLOCK_COUNT__ ) + sizeof( FragmentationSegment ) * ( ( mFragCount + 15 ) / 16 ) ),
            __stack_size( sizeof( AdjacencyTestStack ) + sizeof( FragmentationSegment ) * ( ( mFragCount + 15 ) / 16 ) - sizeof( __BIT_LIST_RAW_DATA_TYPE__ ) ),
            mFragments( createFragments( fstrategy, fsize, mFragCount ) ),
            __probabilities( fstrategy.getFragHitProbability( mFragments, mFragCount ) ) {}
        
        ~FragmentationTransformation () {
             delete[] mFragments;
             delete[] __probabilities;
         }
        
        FragmentationData convert ( const __BIT_LIST_RAW_DATA_TYPE__ * const map ) const {
            FragmentationData result( mFragCount );
            
            for ( uint_fast16_t findex = 0; findex < mFragCount; ++findex ) {
                result.set_frag_value( mFragments[ findex ].getCommonPopCnt( map ), findex );
            } 
            
            return result;
        }
        
        inline FragmentationData convert ( const BitList & map ) const {
            return convert( map.data() );
        } 

        inline void apply ( EdgeCandidate & cand ) const { 
            cand.stack()->load_reg_1( ( mFragCount + 15 ) / 16 );
            apply( cand.adj_cand().data(), cand.stack()->start_ptr() );
        }
        
        inline void apply ( const NarrowingCandidate * const cand, AdjacencyTestStack * const stack ) const { 
            stack->load_reg_1( ( mFragCount + 15 ) / 16 );
            apply( cand->data(), stack->start_ptr() );
        }
        
        unsigned int getPopulationCount ( const BitList & map, const unsigned int findex ) const {
            return map.getCommonPopCnt( mFragments[ findex ] );
        }
        
        inline uint_fast16_t getFragmentSize () const {
            return mFragSize;
        }
        
        inline uint_fast16_t getFragmentCount () const {
            return mFragCount;
        }

        uint_fast16_t getAdjacencyCandidateSize () const {
            return __adjacency_candidate_size;
        }
        
        uint_fast16_t getAdjacencyCandidateOffset () const {
            return __adjacency_candidate_size % sizeof( FragmentationSegment );
        }
        
        uint_fast16_t getAdjacencyStackSize () const {
            return __stack_size;
        }
        
        double_t const * const & probabilities () const {
            return __probabilities;
        }
        
        const std::string toString () const;
    };
}    

#endif /* __DD_FRAGMENTATION_TRANSFORMATION__ */
