/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_INTERVAL__
#define __DD_INTERVAL__

#include <string>
#include <sstream>

namespace addibit {

    class Interval {
        
        private:
        uint_fast32_t mLowerBound;
        uint_fast32_t mUpperBound;

        public:
        Interval () :
            mLowerBound( 1 ),
            mUpperBound( 0 ) {}

        Interval ( const uint_fast32_t lb,
                   const uint_fast32_t ub ) :
            mLowerBound( lb ),
            mUpperBound( ub ) {}

        Interval ( const Interval & other ) :
            mLowerBound( other.mLowerBound ),
            mUpperBound( other.mUpperBound ) {}

        uint_fast32_t & lb () {
            return mLowerBound;
        }

        const uint_fast32_t & lb () const {
            return mLowerBound;
        }

        uint_fast32_t & ub () {
            return mUpperBound;
        }

        const uint_fast32_t & ub () const {
            return mUpperBound;
        }

        bool empty () const {
            return lb() > ub();
        }

        bool pointed () const {
            return lb() == ub();
        }

        uint_fast32_t range () const {
            return ( 1 + ub() ) - lb();
        }

        bool contains ( const uint_fast32_t value ) const {
            return value >= lb() && value <= ub();
        }

        void update ( const uint_fast32_t value ) {

            if ( unlikely( empty() ) ) {
                lb() = value;
                ub() = value;
            } else if ( value > ub() ) {
                ub() = value;
            } else if ( value < lb() ) {
                lb() = value;
            }
        }

        void update ( const Interval & value ) {
            update( value.lb() );
            update( value.ub() );
        }

        void operator= ( const Interval & rhs ) {
            mLowerBound = rhs.mLowerBound;
            mUpperBound = rhs.mUpperBound;
        }

        bool operator > ( const uint_fast32_t value ) const {
            return lb() > value;
        }

        bool operator >= ( const uint_fast32_t value ) const {
            return lb() >= value;
        }

        bool operator <= ( const uint_fast32_t value ) const {
            return ub() <= value;
        }

        bool operator < ( const uint_fast32_t value ) const {
            return ub() < value;
        }

        void reset () {
            mLowerBound = 1;
            mUpperBound = 0;
        }

        const std::string toString () const {
            std::stringstream result;

            if ( empty() ) {
                result << "[empty]";
            } else {
                result << "[" << lb() << ".." << ub() << "]";
            }

            return result.str();
        }
    };
}

#endif // __DD_INTERVAL__
