/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "DDMatrix.h"
#include "DDUtilsFactory.h"
#include "DDSortCriterionMatrixWrapper.h"

#include "DDLogging.h"

#include <sstream>
#include <algorithm>

using namespace addibit;

Matrix::Matrix () :
    __equalities( 0 ),
    __strict( 0 ) {}

Matrix::Matrix ( const uint_fast16_t rows, const uint_fast16_t columns ) :
    __equalities( 0 ),
    __strict( 0 ) {

	for ( uint_fast16_t row_index = 0; row_index < rows; row_index++ ) {
		const CalculationVector row( columns );
		mData.push_back( row );
		mMapping.push_back( row_index );
	}	
}

Matrix::Matrix ( const uint_fast16_t rows, const uint_fast16_t columns, const std::vector<uint_fast16_t> & mapping ) :
	mMapping( mapping ),
    __equalities( 0 ),
    __strict( 0 ) {

	for ( uint_fast16_t row_index = 0; row_index < rows; row_index++ ) {
		const CalculationVector & row = CalculationVector( columns );
		mData.push_back( row );
	}		
}

Matrix::Matrix ( const Matrix & other ) :
	mData( other.mData ),
	mMapping( other.mMapping ),
    __equalities( other.__equalities ),
    __strict( other.__strict ) {}

void Matrix::insert ( const CalculationVector& row ) {
	mMapping.push_back( mData.size() );
	mData.push_back( row );
}

Matrix Matrix::operator * ( const Matrix & rhs ) const {
	Matrix result = *this;
	
	result *= rhs;

	return result;
}

void Matrix::operator *= ( const Matrix& matrix ) {

	CalculationVector buffer( matrix.getColumns() );

	for ( uint_fast16_t row = 0; row < getRows(); row++ ) {
		
		for ( uint_fast16_t column = 0; column < matrix.getColumns(); column++ ) {

			buffer.setValue( column, 0 );

			for ( uint_fast16_t index = 0; index < getColumns(); index++ ) {
				if ( isNonZeroAt( row, index ) && matrix.isNonZeroAt( index, column ) ) {

					UtilsFactory::CALC_UTILS.loadRegisterT( getValue( row, index ) );
					UtilsFactory::CALC_UTILS.multiplyRegisterT( matrix.getValue( index, column ) );
					buffer.add( column, UtilsFactory::CALC_UTILS.getRegisterT() );

				}
			}
		}	
		getRow( row ) = buffer; 
	}
}

void Matrix::operator = ( const Matrix& matrix ) {
	mData = matrix.mData;
	mMapping = matrix.mMapping;
}

Matrix Matrix::getSubMatrix ( const uint_fast16_t beginRow, const uint_fast16_t endRow ) const {
	Matrix result;	

	for ( uint_fast16_t index = beginRow; index < endRow; index++ ) {
		result.insert( getRow( index ) );
	}

	return result;
}

void Matrix::mark ( const std::vector<uint_fast16_t> & strict, const std::vector<uint_fast16_t> & eq ) {
    std::vector<uint_fast16_t> strict_mapping;
    std::vector<uint_fast16_t> eq_mapping;
    
    for ( auto index : strict ) {
        strict_mapping.push_back( mMapping[ index ] );
        mMapping[ index ] = mData.size();
    }
    
    for ( auto index : eq ) {
        eq_mapping.push_back( mMapping[ index ] );
        mMapping[ index ] = mData.size();
    }
    
    mMapping.erase( std::remove( mMapping.begin(), mMapping.end(), mData.size() ), mMapping.end() );
    mMapping.insert( mMapping.begin(), eq_mapping.begin(), eq_mapping.end() );
    mMapping.insert( mMapping.end(), strict_mapping.begin(), strict_mapping.end() );
    __equalities = eq.size();
    __strict = strict.size();
}


void Matrix::sort ( const SortCriterion & criterion ) {
    SortCriterionMatrixWrapper comparator( criterion, mData );
    std::sort( mMapping.begin() + __equalities, mMapping.end() - __strict, comparator );	
}

const std::string Matrix::getMapping () const {
	std::stringstream output;

	output << "( ";
	for ( unsigned long i = 0; i < getRows() - 1; i++ ) {
		output << "\"" << mMapping[i] << "\" ";
	}
	output << "\"" << mMapping[ getRows()-1 ] << "\" )";

	return output.str();
}

const std::string Matrix::toString () const {
	std::stringstream output;

	for ( unsigned long i = 0; i < getRows() - 1; i++ ) {
		output << getRow( i ).toString() << std::endl;
	}
	output << getRow( getRows() - 1 ).toString();

	return output.str();
}

const std::string Matrix::toPortaFormat( const mpz_class & val, bool first = false ) const {
	std::stringstream output;
    
    if ( val > 0 && !first ) {
        output << "+";
    } else if ( val == -1 ) {
        output << "-";
    }
    if ( val != 1 && val != -1 ) {
        output << val;
    }

    return output.str();
}

const std::string Matrix::to4ti2Format () const {
    std::stringstream output;

    output << getRows() << " " << getColumns() << std::endl;

    for ( uint_fast16_t r_index = 0; r_index < getRows(); r_index++ ) {
        const CalculationVector & row = getRow( r_index );

        for ( uint_fast16_t c_index = 0; c_index < getColumns() - 1; c_index++ ) {
            output << row[ c_index ] << " ";
        }     
        output << row[ getColumns() - 1 ] << std::endl;
    }

    return output.str();
}

const std::string Matrix::to4ti2RelFormat () const {
    std::stringstream output;

    output << "1 " << getRows() << std::endl;

    for ( uint_fast16_t r_index = 0; r_index < getRows() - 1; r_index++ ) {
        output << "1 ";
    }
    output << "1" << std::endl;

    return output.str();
}

const std::string Matrix::to4ti2SignFormat () const {
    std::stringstream output;

    output << "1 " << getColumns() << std::endl;

    for ( uint_fast16_t c_index = 0; c_index < getColumns() - 1; c_index++ ) {
        output << "0 ";
    }
    output << "0" << std::endl;

    return output.str();    
}

const std::string Matrix::toPortaFormat () const {
	std::stringstream output;

    output << "DIM = " << getColumns() << std::endl
           << "INEQUALITIES_SECTION" << std::endl;

    for ( uint_fast16_t r_index = 0; r_index < getRows(); r_index++ ) {
        const CalculationVector & row = getRow( r_index );
        uint_fast16_t c_index = 0;
        
        while ( !row.isNonZeroAt( c_index ) ) {
            c_index++;
        }

        output << toPortaFormat( row[ c_index ], true ) << "x" << ( c_index + 1 );
        c_index++;

        for ( ; c_index < getColumns(); c_index++ ) {
            if ( row.isNonZeroAt( c_index ) ) {
                output << toPortaFormat( row[ c_index ] ) << "x" << ( c_index + 1 );
            }
        }     
        output << ">=0" << std::endl;
    }

    output << "END";
    
    return output.str();
}
