/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_REPRESENTATION_VECTOR__
#define __DD_REPRESENTATION_VECTOR__

#include "DDVector.h"
#include "DDCalculationVector.h"

#include <vector>
#include <string.h>
#include <sstream>

#include <gmp.h>
#include <gmpxx.h>

namespace addibit {

	class RepresentationVector : public Vector {

		private:
		uint_fast16_t * indexToDataMap;
		uint_fast16_t dim;

		public:
		RepresentationVector ( const uint_fast16_t dimension ) :
			Vector(),
			dim( dimension ) {
		
			indexToDataMap = new uint_fast16_t[ dimension ];

			for ( uint_fast16_t i = 0; i < dimension; i++ )
				indexToDataMap[ i ] = 0;
		}

		RepresentationVector ( const RepresentationVector& other ) :
			Vector( other ),
			dim( other.dim ) {

			indexToDataMap = new uint_fast16_t[dim];

			for ( uint_fast16_t i = 0; i < dim; i++ ) {
				indexToDataMap[i] = other.indexToDataMap[i];
            }
		}

		virtual ~RepresentationVector () {
			delete[] indexToDataMap;
		}

		uint_fast16_t getDimension () const {
			return dim;
		}
		
		bool isNonZeroAt ( const uint_fast16_t index ) const {
			return indexToDataMap[index] != 0;
		}

		uint_fast16_t getNonZeroCount () const {
			return data.size();
		}

		const mpz_class& getValue ( const uint_fast16_t index ) const {
			return data[indexToDataMap[index] - 1];
		}

		void setValue (	const uint_fast16_t index, const mpz_class& value ) {
			if ( isNonZeroAt( index ) ) {
				data[indexToDataMap[index] - 1] = value;
			} else if ( value != 0 ) {
				data.push_back( value );
				indexToDataMap[index] = data.size();
			}
		}

		void setValue (	const uint_fast16_t index, const int_fast64_t value ) {
			mpz_class big_int_value = value;
			setValue( index, big_int_value );
		}

		void add ( const uint_fast16_t index, const mpz_class& increment ) {
			if ( isNonZeroAt( index ) ) {
				data[indexToDataMap[index] - 1] += increment;
			} else {
				setValue( index, increment );
			}
		}

		void divide ( const unsigned long index, const mpz_class& divisor ) {
			if ( isNonZeroAt( index ) ) {
				data[indexToDataMap[index] - 1] /= divisor;
			} 
		}

		void multiply ( const uint_fast16_t index, const mpz_class & scalar ) {
			if ( isNonZeroAt( index ) ) {
				data[indexToDataMap[index] - 1] *= scalar;
			}
		}

		void subtract ( const uint_fast16_t index, const mpz_class& subtractor ) {
			if ( isNonZeroAt( index ) ) {
				data[indexToDataMap[index] - 1] =- subtractor;
			} 
		}

		void negate ( const uint_fast16_t index ) {
			if ( isNonZeroAt( index ) ) {
				data[indexToDataMap[index] - 1] = -data[indexToDataMap[index] - 1];
			} 
		}

		void negate () {
			for ( uint_fast16_t index = 0; index < data.size(); index++ ) {
				data[index] = -data[index];
			} 
		}

		const mpz_class& operator[] ( const uint_fast16_t index ) const {
			return getValue( index );
		}

		bool operator == ( const RepresentationVector& vector ) const {
			
			for ( uint_fast16_t index = 0; index < getDimension(); index++ ) {
				if ( ( vector.isNonZeroAt( index ) xor isNonZeroAt( index ) ) || ( isNonZeroAt( index ) && vector[index] != getValue( index )) ) {
					return false;
				}
			}

			return true;
		}

		void calcScalarProduct ( const RepresentationVector & vector, mpz_class & result ) const {
			result = 0;

			for ( uint_fast16_t index = 0; index < getDimension(); index++ ) {
				if ( isNonZeroAt( index ) && vector.isNonZeroAt( index ) ) {
					result += vector[index] * getValue( index );
				}
			}		
		}

		const mpz_class operator * ( const CalculationVector& vector ) const {
			mpz_class result;
			
			calcScalarProduct( vector, result );

			return result;
		}

		void calcScalarProduct ( const CalculationVector& vector, mpz_class & result ) const {
			result = 0;

			for ( uint_fast16_t index = 0; index < getDimension(); index++ ) {
				if ( isNonZeroAt( index ) ) {
					result += vector[index] * getValue( index );
				}
			}		
		}

        void clear () {
            memset( indexToDataMap, 0, dim * sizeof( uint_fast16_t ) );
            Vector::clear();
        }

		const std::string toString () const {
	        std::stringstream output;
            const uint_fast16_t last = getDimension();

	        for ( uint_fast16_t i = 0; i < last; i++ ) {
		        output << ( isNonZeroAt( i ) ? getValue( i ) : 0 ) << " ";
	        }

	        return output.str();
        }
	};
}

#endif // __DD_REPRESENTATION_VECTOR__
