/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "DDTreeFactory.h"

#include "DDSingleBitSplitVisitor.h"
#include "DDBinaryFragmentedSingleBitSplitVisitor.h"
#include "DDBinaryFragmentedTwoBitsSplitVisitor.h"
#include "DDVantagePointVisitor.h"
#include "DDVantagePointTransformation.h"
#include "DDFragmentationSplitVisitor.h"

#include "DDVantagePointTreeNodeCondition.h"
#include "DDTreeCreator.h"
#include "DDLeafTreeCreator.h"

using namespace addibit;

AbstractTreeCreator * TreeFactory::getBitPatternTreeCreator ( const TreeOptions & options, const BitSelectStrategy & strategy ) const {
    AbstractTreeCreator * creator;
	
	if ( options.fbpt_fragmentation() == 2 ) {	
	    BinaryFragmentedSingleBitSplitVisitor * visitor = new BinaryFragmentedSingleBitSplitVisitor( options.max_tree_depth(), options.leaf_capacity(), strategy, 2, GLOBAL_BLOCKS );   
	    creator = new TreeCreator<TwoBitsFragmentationCondition, TwoBitsFragmentationCondition, BinaryFragmentedSingleBitSplitVisitor, EmptyTransformation>( visitor, new EmptyTransformation(), options.min_elements() );
	} else {
	    SingleBitSplitVisitor * visitor = new SingleBitSplitVisitor( options.max_tree_depth(), options.leaf_capacity(), strategy );
	    creator = new TreeCreator<TreeNodeCondition, TreeNodeCondition, SingleBitSplitVisitor, EmptyTransformation>( visitor, new EmptyTransformation(), options.min_elements() );
	}

	return creator;
}

AbstractTreeCreator * TreeFactory::getExtendedBitPatternTreeCreator( const TreeOptions & options, const BitSelectStrategy & strategy ) const {
    SingleBitSplitVisitor * visitor = new SingleBitSplitVisitor( options.max_tree_depth(), options.leaf_capacity(), strategy );
    return new TreeCreator<ExtendedBitPatternCondition, ExtendedBitPatternCondition, SingleBitSplitVisitor, EmptyTransformation>( visitor, new EmptyTransformation(), options.min_elements() );
}

AbstractTreeCreator * TreeFactory::getVantagePointTreeCreator ( const TreeOptions & options, const BitSelectStrategy & strategy ) const {

    const VantagePointTransformation * transformation = new VantagePointTransformation( options.max_tree_depth(), strategy );
	VantagePointVisitor * visitor = new VantagePointVisitor( options.max_tree_depth(), options.leaf_capacity(), transformation->static_vp(), options.vantage_point_distance_coefficient() );
	return new TreeCreator<VantagePointTreeNodeCondition, VantagePointTreeNodeCondition, VantagePointVisitor, VantagePointTransformation>( visitor, transformation, options.min_elements() );
}
        
AbstractTreeCreator * TreeFactory::getFragmentedBitPatternTreeCreator ( const TreeOptions & options, const BitSelectStrategy & strategy ) const {

    if ( options.fbpt_fragmentation() == 2 ) {
        BinaryFragmentedTwoBitsSplitVisitor * visitor = new BinaryFragmentedTwoBitsSplitVisitor( options.max_tree_depth(), options.leaf_capacity(), GLOBAL_BLOCKS );   
	    return new TreeCreator<TwoBitsFragmentationCondition, TwoBitsFragmentationCondition, BinaryFragmentedTwoBitsSplitVisitor, EmptyTransformation>( visitor, new EmptyTransformation(), options.min_elements() );      
    } else {
        return getBitPatternTreeCreator( options, strategy );
    }
}

AbstractTreeCreator * TreeFactory::getFragmentationTreeCreator ( const TreeOptions & options, const BitSelectStrategy & strategy ) const {
	AbstractTreeCreator * creator;
	
	const FragmentationTransformation * transformation = new FragmentationTransformation( options.fbpt_fragmentation(), strategy );
    FragmentationSplitVisitor * visitor = new FragmentationSplitVisitor( options.max_tree_depth(), options.leaf_capacity(), *transformation, options.fbpt_fragmentation() );
    
    creator = new TreeCreator<FragmentationCondition, FragmentationCondition, FragmentationSplitVisitor, FragmentationTransformation>( visitor, transformation, options.min_elements() );
    
    return creator;
}

AbstractTreeCreator * TreeFactory::getTreeCreator ( const TreeOptions & options, const BitSelectStrategy & strategy ) const {
	
	const TreeType tree_type = options.tree_type();

	switch ( tree_type ) {
	
		case BINARY_STATIC_VANTAGE_POINT:
		return getVantagePointTreeCreator( options, strategy );
		
		case FRAGMENTED_BIT_PATTERN_TREE:
		return getFragmentedBitPatternTreeCreator( options, strategy );
		
		case FRAGMENTATION_TREE:
		return getFragmentationTreeCreator( options, strategy );
		
        case BIT_PATTERN_TREE:
        return getBitPatternTreeCreator( options, strategy );
    
        case EXTENDED_BIT_PATTERN_TREE:
        return getExtendedBitPatternTreeCreator( options, strategy );
        
		default:
		return new LeafTreeCreator();
	}
}
