/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <deque>
#include <iostream>
#include <stdlib.h>
#include <memory.h>

#include "DDConstants.h"
#include "DDInput.h"
#include "DDOutput.h"
#include "DDStats.h"

#include "DDOutputVector.h"
#include "DDRepresentationVector.h"
#include "DDExtremeRayUncompressor.h"

#include "DDLogging.h"

#include "DDOptions.h"
#include "DDCacheFactory.h"

using namespace addibit;

template <class RAY_TYPE>
double_t calc_degeneracy ( const std::deque<std::shared_ptr<const RAY_TYPE > > * extreme_rays,
                           const uint_fast16_t threads ) {
                                
    double_t degree = 0;
    uint_fast64_t number = 0;
    
    for ( uint16_t t_id = 0; t_id < threads; t_id++ ) {
        number += extreme_rays[ t_id ].size();
        
        for ( auto & ray : extreme_rays[ t_id ] ) {
            degree += ray->deg_degree();
        }
    }  
    
    return degree / number;
}

template <class RAY_TYPE>
uint_fast64_t count_extreme_rays ( const std::deque<std::shared_ptr<const RAY_TYPE > > * extreme_rays,
                                   const uint_fast16_t threads ) {
    uint_fast64_t number = 0;

    for ( uint16_t t_id = 0; t_id < threads; t_id++ ) {
        number += extreme_rays[ t_id ].size();
    }  

    return number;
}

template <class RAY_TYPE, class VECTOR, typename ... Args>
uint_fast64_t write_extreme_rays ( const char * file,
                                   const Matrix & system,
                                   const std::deque<std::shared_ptr<const RAY_TYPE > > * extreme_rays,
                                   const uint_fast16_t threads,
                                   Args ... args ) {
                              
    Output output( file );
    ExtremeRayUncompressor decompressor( system.getRows(), system.getColumns() );
    VECTOR current( args... );
    uint_fast64_t number = count_extreme_rays<RAY_TYPE>( extreme_rays, threads );
    
    output << "begin\n"
           << " " << number << " " << system.getColumns() << " integer\n";

    for ( uint16_t t_id = 0; t_id < threads; t_id++ ) {
        for ( auto & ray : extreme_rays[ t_id ] ) {
            decompressor.calcExtremeRay( system, ray.get(), current );
            output.write( current );
            current.clear();
        }
    }
    
    output << "end";    
    
    return number;
}

template <class RAY_TYPE, class CACHE_ENTRY_TYPE, class GENERATOR>
int generate ( const char * problem ) {

	START_TIMER( OVERALL_TIME )

    Options & options = Options::current();
	Input input ( problem );
	const Matrix & system = input.getSystem();
    if ( input.ordering() != NO_ORDERING )
        options.input_ordering() = options.input_ordering();
    
    const uint_fast16_t threads = 1 << options.thread_exponent();
    std::deque<std::shared_ptr<const RAY_TYPE> > * extreme_rays = new std::deque<std::shared_ptr<const RAY_TYPE> >[ threads ];

    init_logging();
    init_constants( BitList::getBlocks( system.getRows() ), system.getColumns() - 2, system.getRows() );

    DD_OUTPUT( options.toString() )

    DD_OUTPUT ( "-----------------------------------------------------------------------------------" )
    DD_OUTPUT ( "|                                  COMPUTATION                                    |" )
    DD_OUTPUT ( "-----------------------------------------------------------------------------------" )


    GENERATOR generator( system );

#ifdef __STATS_OUTPUT__
    if ( options.plot_required() ) {
        OPEN_LOG( COMPLEXITY_PLOT )
        OPEN_LOG( VER_BIT_DISTRIBUTION_PLOT )
    }
#endif

    const uint_fast16_t list_like_entries = system.getRows() - ( system.getColumns() + options.thread_exponent() + system.strict() + 1 );
    CacheFactory<RAY_TYPE, CACHE_ENTRY_TYPE> factory( extreme_rays, threads );
    CacheEntryNode<RAY_TYPE, CACHE_ENTRY_TYPE> * head = factory.create_cache( system.getColumns(), list_like_entries, system.getRows() - system.strict() );
        
    DUMP_STRING ( "sorted_matrix.txt", system.toString() );
    DUMP_STRING ( "cache.dot", head->toGraphVizString() ); 

    START_TIMER( ITERATIONS_TIME )
    generator.init( head );
    generator();
    STOP_TIMER( ITERATIONS_TIME )

    const std::string res_file = std::string( problem ) + ".out";
    uint_fast64_t number;

    switch ( options.output_format() ) {
        case INTEGER_VEC:
            number = write_extreme_rays<RAY_TYPE, RepresentationVector, uint_fast16_t>( res_file.c_str(), system, extreme_rays, threads, system.getColumns() );
            break;
        case DOUBLE_VEC:
            number = write_extreme_rays<RAY_TYPE, OutputVector<double>, uint_fast16_t>( res_file.c_str(), system, extreme_rays, threads, system.getColumns() );
            break;
        case ACTIVE_SET:
            number = write_extreme_rays<RAY_TYPE, BitList, unsigned long *>( res_file.c_str(), system, extreme_rays, threads, __BIT_LIST_ALLOC_SPACE__ );
            break;
        default: /* COUNT_ONLY */
            number = count_extreme_rays<RAY_TYPE>( extreme_rays, threads );
    }

#ifdef __STATS_OUTPUT__
    if ( options.plot_required() ) {
        DUMP_LOG( COMPLEXITY_PLOT, options.plot_file().c_str() )
        DUMP_LOG( VER_BIT_DISTRIBUTION_PLOT, "ver_bit_distribution.plt" )
    }
#endif

    DD_OUTPUT ( "-----------------------------------------------------------------------------------" )
    DD_OUTPUT ( "Calculated " << number <<  " Extreme Rays" )
#ifdef __STATS_OUTPUT__
    DD_OUTPUT ( "Average degeneracy degree: " << calc_degeneracy( extreme_rays, threads ) )
#endif
    if ( threads > 1 ) { 
        for ( uint_fast16_t t_id = 0; t_id < threads; t_id++ ) {
            DD_OUTPUT ( "Thread " << ( t_id + 1 ) <<  " : " << extreme_rays[ t_id ].size() )
        }
    }
    DD_OUTPUT ( "Narrowing and verification time: " << TIMER_VALUE( ITERATIONS_TIME ) << "ms" )        
    
    STOP_TIMER( OVERALL_TIME )
	DD_OUTPUT( "Overall time: " << TIMER_VALUE( OVERALL_TIME ) << "ms" );

    delete[] extreme_rays;
    Options::release();

	return 0;
}
